@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 2008-2018 Sergey Poznyakoff
+   Copyright (C) 2008-2019 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -310,7 +310,7 @@ parse_acl_line(grecs_locus_t *locus, int allow, dicod_acl_t acl,
 /* ACL verification */
 
 static int
-cmp_group_name(const void *item, void *data)
+cmp_group_name(const void *item, const void *data, void *ignore)
 {
     return strcmp((char*)item, (char*)data);
 }
@@ -366,7 +366,7 @@ _acl_check(struct acl_entry *ent)
 
     if (ent->groups) {
 	result = dico_list_intersect_p(ent->groups, user_groups,
-				       cmp_group_name);
+				       cmp_group_name, NULL);
 	if (!result)
 	    return 0;
     }	
