@@ -1,6 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 1998-2000, 2008, 2010, 2012, 2016-2018 Sergey
-   Poznyakoff
+   Copyright (C) 1998-2019 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -243,6 +242,7 @@ void dicod_capa_register(const char *name, struct dicod_command *cmd,
 int dicod_capa_add(const char *name);
 void dicod_capa_iterate(int (*fun)(const char*, int, void *), void *closure);
 int dicod_capa_flush(void);
+int dicod_capa_is_enabled(const char *name);
 
 /* lang.c */
 void register_lang(void);
