@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 1998-2016 Sergey Poznyakoff
+   Copyright (C) 1998-2019 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -23,7 +23,7 @@ dicod_xidle(dico_stream_t str, int argc, char **argv)
 }
     
 void
-register_xidle()
+register_xidle(void)
 {
     static struct dicod_command cmd[] = {
 	{ "XIDLE", 1, 1, NULL, "report inactivity timeout, reset idle timer",
