@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 2008-2016 Sergey Poznyakoff
+   Copyright (C) 2008-2019 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -14,17 +14,14 @@
    You should have received a copy of the GNU General Public License
    along with GNU Dico.  If not, see <http://www.gnu.org/licenses/>. */
 
-#ifdef HAVE_CONFIG_H
-# include <config.h>
-#endif
+#include <config.h>
 #include <gettext.h>
-#include <configmake.h>
 #include <locale.h>
 
 static int libdico_initialized;
 
 void
-_dico_libi18n_init()
+_dico_libi18n_init(void)
 {
     if (!libdico_initialized) {
 	bindtextdomain(PACKAGE, LOCALEDIR);
